package ASCOPE::Search::Boolean;
use strict;

$ASCOPE::Search::Boolean::VERSION = '1.0';

use IO::File;
use Carp;

# We'll also try to load Memoize
# in an END{} block futher below.

use constant MAX_TRIES => 64;

sub new {
  my $pkg = shift;

  my $self = bless {},$pkg;

  if (! $self->init(@_)) {
    return 0;
  }

  return $self;
}

sub init {
  my $self = shift;

  my $fh = IO::File->new($_[0]);

  if (! $fh) {
    carp "Can't load list to search, $!\n";
    return 0;
  }

  $self->{'__fh'}  = $fh;
  $self->{'__len'} = (stat($_[0]))[7];

  return 1;
}

sub query {
  my $self = shift;

  # For reasons I don't understand, I
  # can't pass \*$self->{'__fh'} without
  # generating errors....
  my $fh = $self->{'__fh'};

  return &_query(\*$fh,$self->{'__len'},$_[0]);
}

sub _query {
  my $fh     = shift;
  my $len    = shift;

  my $lookup = shift;
  my $result = undef;

  my $begin = 0;
  my $end   = $len;

  my $found = 0;
  my $tries = 0;

  while (! $found) {

    # Stopping flogging the horse, Jim!

    if (($begin + 1) == $end) {
      return undef;
    }

    # Just because you're paranoid
    # Don't mean they're not after you.

    if ($tries >= MAX_TRIES) {
      carp
	"Erp! Reached limit for maximum attempts. ".
	"I think something is wrong.\n";
      return undef;
    }

    # Go to half-way spot between $begin
    # and $end and then back up to the start
    # of the line.

    my $guess = int(($begin + $end) /2);
    # print STDERR "[B] $begin [E] $end\n";
    # print STDERR "[$tries] Guess is $guess\n";

    my $pos  = $guess;
    my $stop = 0;

    my $char = "";

    while (! $stop) {
      sysseek($fh,$pos,0);
      sysread ($fh,$char,1);

      if ($char =~ /\n/) {
        $stop = 1;
        $pos  = $pos+2;
      }

      $pos--;

      if (! $pos) {
        $stop = 1;
      }
    }

    # Look forward only enough to determine
    # whether or not we're dealing with an
    # entry that is like our lookup

    $stop  = 0;

    my $line  = undef;
    my $match = 0;

    while (! $stop) {
      sysseek($fh,$pos,0);
      sysread($fh,$char,1);

      $line .= $char;
      # print "[$pos] '$lookup' '$char' '$line'\n";

      if ($line =~ /^$lookup\t/) {
	$pos++;

	$match = 1;
	$stop  = 1;
      }

      elsif ($lookup =~ /^$line/) {
	$pos++;
      }

      # This is not the droid, we're looking for.
      # The only question now is whether to look
      # forwards or backwards.

      else {

	# print "Compare '$lookup' and '$line'\n";

	if ([sort ($lookup,$line)]->[0] eq $lookup) {
	  # print "...before ($begin | $end)\n";
	  $end = $guess;
	}

	else {
	  # print "...after ($begin | $end)\n";
	  $begin = $guess;
	}

	$stop = 1;
      }
    }

    $tries ++;
    next unless ($match);

    # Go forward until the end of
    # the line. Return whatever we
    # find.

    $match = undef;
    $stop  = 0;

    while (! $stop) {
      sysseek($fh,$pos,0);
      sysread ($fh,$char,1);

      $pos++;

      if ($char =~ /\n/) {
	$stop  = 1;
	$found = 1;

	last;
      }

      # Check for dangling tabs

      elsif ($char =~ /\t/) {
	next;
      }

      else {
	$result .= $char;
      }

    }

  }

  # Oop oop oop

  return ASCOPE::Search::Boolean::result->new([$result,$lookup]);
}

END {
  eval "require Memoize";
  if (! $@) { Memoize::memoize("_query"); }
}

package ASCOPE::Search::Boolean::result;

use overload q("") => sub { shift->[0] };

sub new   { return bless $_[1],$_[0]; }
sub query { shift->[1]; }

return 1;

__END__

=head1 NAME

ASCOPE::Search::Boolean - Aaron's generic boolean search class.

=head1 SYNOPSIS

  my $search = ASCOPE::Search::Boolean->new($list);
  my $result = $search->query("foobar");

  print $result;

  # prints:
  # a-oooooooo-ga!

  print $result->query();

  # prints:
  # foobar

  print ref($result);

  # prints:
  # ASCOPE::Search::Boolean::result

=head1 DESCRIPTION

This is Aaron's generic boolean search class.

It performs lookups on files containing I<foo\tbar\n> style lists
using a boolean search.

Hashes in Perl are super cool but they come at the expense of being
pretty big in terms of memory. And if you have a long  list  of
foo->bar query pairs then you also have a *very* big hash.

Some people take their lists and write them to a file and simply loop
over every line which is, overall, better in terms of memory usage. Except
if your file is very long or your lookup is all the way at the end.

Ultimately, it depends on the size of your list. If you've only got 4-5
thingies then a hash is fine.  But if you've got a couple hundred entries
and you can imagine it growing using a boolean search is worth the trouble
as it is very fast and doesn't have the same overhead.

=head1 PACKAGE METHODS

=head2 __PACKAGE__->new($path)

Where I<$path> is the path to a file containing your foo->bar pairs.
see also: examples/diacratics.txt

Returns an object. Woot!

=head1 OBJECT METHODS

=head2 $obj->query($str)

Returns under if no match was found.

If a match is found, returns an I<ASCOPE::Search::Boolean::result> object.

A I<::result> object's I<""> operator is overloaded and will return the
result of your query.

The object defines a single I<query> method which will return the text used
to retrieve the result itself.

=head1 VERSION

1.0

=head1 DATE

$Date: 2003/03/12 23:16:41 $

=head1 AUTHOR

Aaron Straup Cope

=head1 NOTES

=over 4

=item *

The package will try to load I<Memoize.pm> for caching lookups
but it will not fail if it is not installed.

=back

=head1 LICENSE

Copyright (c) 2003, Aaron Straup Cope. All Rights Reserved.

This is free software, you may use it and distribute it under
the same terms as Perl itself.

=cut
